function [model,sample2node] = forestTrain(X, Y, opts)
    % X is NxD, where rows are data points
    % for convenience, for now we assume X is 0 mean unit variance. If it
    % isn't, preprocess your data with
    %
    % X= bsxfun(@rdivide, bsxfun(@minus, X, mean(X)), var(X) + 1e-10);
    %
    % If this condition isn't satisfied, some weak learners won't work out
    % of the box in current implementation.
    %
    % Y is discrete Nx1 vector of labels
    % model can be plugged into forestTest()
    %
    % decent default opts are:
    % opts.depth= 9;
    % opts.numTrees= 100; %(but more is _ALWAYS_ better, monotonically, no exceptions)
    % opts.numSplits= 5;
    % opts.classifierID= 2
    % opts.minscount = 0; %minimum number of samples in each branch for
    % split to be accepted
    % opts.minspn = 0; %minimum number of samples in a node to split it
    % opts.classifierCommitFirst= true;
    % opts.saveIds = false;
    %
    % which means use depth 9 trees, train 100 of them, use 5 random
    % splits when training each weak learner. The last option controls
    % whether each node commits to a weak learner at random and then trains
    % it best it can, or whether each node tries all weak learners and
    % takes the one that does best. Empirically, it appears leaving this as
    % true (default) gives slightly better looking results.
    %
    
    numTrees= 100;
    verbose= false;
    
    warning('Using new (memsave) version of the forest');
    
    if(size(Y,1)<size(Y,2)), Y = Y'; end;
    if(size(Y,2)~=1), error('Y should be a column vector'); end;
    if(size(X,1)~=size(Y,1)), error('X and Y sizes are different'); end;
    
    if nargin < 3, opts= struct; end
    if isfield(opts, 'numTrees'), numTrees= opts.numTrees; end
    if isfield(opts, 'verbose'), verbose= opts.verbose; end
    if ~isfield(opts,'minscount'), opts.minscount = 0; end %no minimum number of samples in the node
    if ~isfield(opts,'minspn'), opts.minspn = 0; end %minimum number of samples in each branch for split to take place
    if ~isfield(opts, 'saveIds'), opts.saveIds = false; end
    if nargout>1, sample2node = zeros(size(X,1),opts.numTrees); end
    
    %create indexing structure
    opts.index = createIndex();
    
    mlength = 0;
    for i=1:length(opts.classifierID)
        mlength = max(mlength,opts.index{opts.classifierID(i)}.length);
    end
    
    opts.mlength = mlength;
    opts.mwIDidx = 1;
    opts.nosplitidx = 2;
    
    model.mlength = opts.mlength;
    model.mwIDidx = opts.mwIDidx; 
    model.nosplitidx = opts.nosplitidx;

    treeModels= cell(1, numTrees);
    for i=1:numTrees
        
        if (nargout>1)
            [treeModels{i},s2ni] = treeTrain(X, Y, opts);
            sample2node(:,i) = s2ni;
        else
            treeModels{i} = treeTrain(X, Y, opts);
        end
        
        % print info if verbose
        if verbose
            p10= floor(numTrees/10);
            if mod(i, p10)==0 || i==1 || i==numTrees
                fprintf('Training tree %d/%d...\n', i, numTrees);
            end
        end
    end
    
    model.dim = size(X,2);
    model.treeModels = treeModels;
end